package tools.descartes.dql.connector.simQPN;

import java.math.BigDecimal;
import java.util.LinkedList;
import java.util.List;

import tools.descartes.dql.models.mapping.domain.DecimalResult;
import tools.descartes.dql.models.mapping.domain.DoF;
import tools.descartes.dql.models.mapping.domain.DomainFactory;
import tools.descartes.dql.models.mapping.domain.Probe;
import tools.descartes.dql.models.mapping.domain.Resource;
import tools.descartes.dql.models.mapping.domain.Service;

public class EntityMappingUtil {

	public static Service createService(String name) {
		final Service service = DomainFactory.eINSTANCE.createService();
		service.setIdentifier(name);
		service.setAlias(name);
		return service;
	}

	public static Resource createResource(String name) {
		final Resource resource = DomainFactory.eINSTANCE.createResource();
		resource.setIdentifier(name);
		resource.setAlias(name);
		return resource;
	}
	
	public static DecimalResult createDecimalResult(String metricName, Double value, double accuracy) {
		return createDecimalResult(metricName, value, accuracy, true);
	}
	
	public static DecimalResult createDoFParameter(String parameterName, Double value) {
		final DecimalResult result = DomainFactory.eINSTANCE.createDecimalResult();
		result.setMetricName(parameterName);
		result.setValid(true);
		result.setAccuracy(BigDecimal.valueOf(1.0));
		result.setValue(BigDecimal.valueOf(value));
		return result;
	}
	
	public static DecimalResult createDecimalResult(String metricName, Double value, double accuracy, boolean valid) {
		final DecimalResult result = DomainFactory.eINSTANCE.createDecimalResult();
		result.setMetricName(metricName);
		result.setValid(valid);
		if (valid)
			result.setAccuracy(BigDecimal.valueOf(accuracy));
		else
			result.setAccuracy(BigDecimal.valueOf(0.0));
		try {
		result.setValue(BigDecimal.valueOf(value));
		} catch (Exception e) {
			System.out.println();
		}
		return result;
	}
	
	public static Probe createProbe(String name) {
		Probe p = DomainFactory.eINSTANCE.createProbe();
		p.setMetricName(name);
		return p;
	}
	
	public static List<Resource> createResources(List<String> resourceNames) {
		List<Resource> resources = new LinkedList<Resource>();
		for (String resourceName: resourceNames)
			resources.add(createResource(resourceName));
		return resources;
	}
	public static List<Service> createServices(List<String> serviceNames) {
		List<Service> services = new LinkedList<Service>();
		for (String serviceName: serviceNames)
			services.add(createService(serviceName));
		return services;
	}

	public static List<DoF> createDoFs(List<String> dofNames) {
		List<DoF> dofs = new LinkedList<DoF>();
		for (String dofName: dofNames)
			dofs.add(createDof(dofName));
		return dofs;
	}

	public static DoF createDof(String dofName) {
		DoF dof = DomainFactory.eINSTANCE.createDoF();
		dof.setIdentifier(dofName);
		dof.setAlias(dofName);
		return dof;
	}
}
