package extension;

import org.dom4j.Document;
import org.dom4j.DocumentHelper;
import org.dom4j.Element;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.util.EObjectContainmentEList;

import tools.descartes.dql.models.mapping.domain.Aggregate;
import tools.descartes.dql.models.mapping.domain.DoF;
import tools.descartes.dql.models.mapping.domain.DomainPackage;
import tools.descartes.dql.models.mapping.domain.Resource;
import tools.descartes.dql.models.mapping.domain.Service;
import tools.descartes.dql.models.mapping.filter.Conjunction;
import tools.descartes.dql.models.mapping.mapping.EntityMapping;
import tools.descartes.dql.models.mapping.mapping.impl.MappingImpl;

public class ExtendedEntityMapping extends MappingImpl implements EntityMapping, Cloneable{
	private String doFLocation = null;
	private EList<Aggregate> aggregates = new EObjectContainmentEList<Aggregate>
			(Aggregate.class, this, DomainPackage.AGGREGATE);
	private EList<DoF> dofs = new EObjectContainmentEList<DoF>
			(DoF.class, this, DomainPackage.DO_F);
	private EList<Resource> resources = new EObjectContainmentEList<Resource>
			(Resource.class, this, DomainPackage.RESOURCE);
	private EList<Service> services = new EObjectContainmentEList<Service>
			(Service.class, this, DomainPackage.SERVICE);
	
	
	@Override
	public String getDoFModelLocation() {
		return doFLocation;
	}

	@Override
	public void setDoFModelLocation(String value) {
		this.doFLocation = value;
	}
	@Override
	public EList<Aggregate> getAggregates() {
		return aggregates;
	}

	@Override
	public EList<DoF> getDoF() {
		return dofs;
	}

	@Override
	public EList<Resource> getResources() {
		return resources;
	}

	@Override
	public EList<Service> getServices() {
		return services;
	}
	
	@Override
	public ExtendedEntityMapping clone() {
		ExtendedEntityMapping clone = new ExtendedEntityMapping();
		clone.setDoFModelLocation(doFLocation);
		
		for (DoF dof: dofs) {
			ExtendedDoF edof = ((ExtendedDoF) dof).clone();
			clone.getDoF().add(edof);
		}
		
		for (Resource resource: resources) {
			ExtendedEntity eEn = ((ExtendedEntity) resource).clone();
			clone.getResources().add(eEn);
		}
		
		for (Service service: services) {
			ExtendedEntity eEn = ((ExtendedEntity) service).clone();
			clone.getServices().add(eEn);
		}
		
		return clone;
	}
	
	public Element toXML() {
        Document document = DocumentHelper.createDocument();
        Element extendedEntity = document.addElement("ExtendedEntityMapping")
        		.addAttribute("doFModelLocation", doFLocation);

        Element dofContainer = extendedEntity.addElement("DoFs");
		for (DoF dof: dofs) {
			ExtendedDoF edof = ((ExtendedDoF) dof).clone();
			dofContainer.add(edof.toXML());
		}

        Element resContainer = extendedEntity.addElement("Resources");
		for (Resource resource: resources) {
			ExtendedEntity eEn = ((ExtendedEntity) resource).clone();
			resContainer.add(eEn.toXML());
		}

        Element serContainer = extendedEntity.addElement("Services");
		for (Service service: services) {
			ExtendedEntity eEn = ((ExtendedEntity) service).clone();
			serContainer.add(eEn.toXML());
		}
		return extendedEntity;
	}
	
	public String toString() {
		return toXML().asXML();
	}
	
	@Override
	public Conjunction getFilter() {
		return null;
	}

	@Override
	public void setFilter(Conjunction value) {
	}


}
