package extension;

import java.util.LinkedList;
import java.util.List;

import org.eclipse.emf.ecore.util.EcoreUtil;

import tools.descartes.dql.connector.ModelStructureQueryConnector;
import tools.descartes.dql.connector.simQPNV2.DQLException;
import tools.descartes.dql.models.mapping.domain.DoF;
import tools.descartes.dql.models.mapping.domain.Entity;
import tools.descartes.dql.models.mapping.domain.Probe;
import tools.descartes.dql.models.mapping.domain.Resource;
import tools.descartes.dql.models.mapping.domain.Service;
import tools.descartes.dql.models.mapping.mapping.EntityMapping;
import tools.descartes.dql.models.mapping.mapping.MappingFactory;
import tools.descartes.dql.models.mapping.mapping.impl.MappingFactoryImpl;

public class ExtensionUtil {

	private static ExtendedEntityMapping extend(EntityMapping entityMapping,
			EntityMapping extensionMap) {
		ExtendedEntityMapping response = new ExtendedEntityMapping();
		response.setDoFModelLocation(entityMapping.getDoFModelLocation());
		response.getAggregates().addAll(entityMapping.getAggregates());
		response.setModelLocation(entityMapping.getModelLocation());
		
		for (Resource originalEntity: entityMapping.getResources()) {
			boolean foundExtension = false;
			for (Resource possibleExtension: extensionMap.getResources()) {
				ExtendedEntity simEn = (ExtendedEntity)(possibleExtension);
				if (simEn.doesExtend(originalEntity)) {
					response.getResources().add((Resource)extend(originalEntity, simEn));
					foundExtension = true;
					break;
				}
			}
			if (!foundExtension)
				throw new DQLException("There is no resource with the name " + originalEntity.getIdentifier());
		}
		
		//can't find a way to avoid having duplicate code
		for (Service originalEntity: entityMapping.getServices()) {
			boolean foundExtension = false;
			for (Service possibleExtension: extensionMap.getServices()) {
				ExtendedEntity simEn = (ExtendedEntity)(possibleExtension);
				if (simEn.doesExtend(originalEntity)) {
					response.getServices().add((Service)extend(originalEntity, simEn));
					foundExtension = true;
					break;
				}
			}
			if (!foundExtension)
				throw new DQLException("There is no service with the name " + originalEntity.getIdentifier());
		}

		//can't find a way to avoid having duplicate code
		for (DoF originalEntity: entityMapping.getDoF()) {
			boolean foundExtension = false;
			for (DoF possibleExtension: extensionMap.getDoF()) {
				ExtendedDoF extenDoF = (ExtendedDoF)(possibleExtension);
				if (extenDoF.doesExtend(originalEntity)) {
					response.getDoF().add(extenDoF.clone());
					foundExtension = true;
					break;
				}
			}
			if (!foundExtension)
				throw new DQLException("There is no service with the name " + originalEntity.getIdentifier());
		}
		
		return response;
	}

	private static Entity extend(Entity originalEntity,
			ExtendedEntity possibleExtension) {
		ExtendedEntity extendedEntity = possibleExtension.clone();
		extendedEntity.setAlias(originalEntity.getAlias());
		
		List<Probe> toAdd = new LinkedList<Probe>();
		for (Probe originalProbe: originalEntity.getProbes()) {
			boolean foundExtension = false;
			for (Probe possibleExtendedProbe: possibleExtension.getProbes()) {
				ExtendedProbe simPro = (ExtendedProbe) possibleExtendedProbe;
				if (simPro.doesExtend(originalProbe)) {
					foundExtension = true;
					toAdd.add(simPro);
				}
			}
			if (!foundExtension)
				throw new DQLException("There is no metric called " + originalProbe.getMetricName() + " for the resource " + originalEntity.getIdentifier());
		}
		extendedEntity.getProbes().clear();
		extendedEntity.getProbes().addAll(toAdd);
		return extendedEntity;
	}

	public static ExtendedEntityMapping extendFullEntityMapping(EntityMapping request,
			ModelStructureQueryConnector modelStructureQueryConnector) {
		MappingFactory mappingFactory = MappingFactoryImpl.init();
		EntityMapping extensionMap = mappingFactory.createEntityMapping();
		extensionMap.setModelLocation(request.getModelLocation());
		extensionMap =  modelStructureQueryConnector.requestEntities(extensionMap);
		extensionMap =  modelStructureQueryConnector.requestProbes(extensionMap);
		return extend(request, extensionMap);
	}
	
	public static ExtendedEntityMapping extendEntitiesOnly(EntityMapping request,
			ModelStructureQueryConnector modelStructureQueryConnector) {
		MappingFactory mappingFactory = MappingFactoryImpl.init();
		EntityMapping extensionMap = mappingFactory.createEntityMapping();
		extensionMap.setModelLocation(request.getModelLocation());
		return extend(request, modelStructureQueryConnector.requestEntities(extensionMap));
	}
}
