package tools.descartes.dql.connector.simQPNV2;

import java.util.HashMap;
import java.util.Map;
import java.util.TreeMap;

import org.dom4j.Element;
import org.dom4j.Node;

import de.tud.cs.qpe.model.NetHelper;
import de.tud.cs.simqpn.ui.model.Configuration;
import extension.ExtendedEntity;
import extension.PlaceType;

public class ConfigurationBuilder {

	private String configName;
	private Double warmupPeriod;
	private Double totalRunLength;
	private String stoppingRule;
	private Double timeBetweenStopChecks;
	private Double secondsBetweenStopChecks;
	private Double accuracy;
	private int verbosity;
	private int timeBeforeInitialHeartBeat;
	private int secondsBetweenHeartBeats;
	private String outputDirectory;
	private Map<ExtendedEntity, Integer> statLevels = new HashMap<ExtendedEntity, Integer>();
	
	public ConfigurationBuilder(String configName) {
		this.configName = configName;
		this.stoppingRule = Configuration.RELATIVE_PRECISION_STOPPING_RULE;
		this.timeBetweenStopChecks = 10.0;
		this.verbosity = 1;
		this.outputDirectory = ".";
		this.totalRunLength = Double.MAX_VALUE;
		this.warmupPeriod = 50000000.0;
		this.secondsBetweenStopChecks = 3.0;
		this.timeBeforeInitialHeartBeat = 1000;
		this.secondsBetweenHeartBeats =  60;
	}
	
	public void addStatLevel(ExtendedEntity place, int level, Element net) {
		if (place.getPlaceType() == PlaceType.QUEUE)
			place = new ExtendedEntity(NetUtil.getPlaceContainingQueue(net,
					place.getXmlId()),
					PlaceType.QUEUEINGPLACE, place.getIdentifier(),
					place.getColorNames());
			
		if (statLevels.get(place) == null || statLevels.get(place) < level) {
			statLevels.put(place, level);
		}
	}
	
	//TODO copy examples into project
	public void writeConfiguration(Element net) {
		IdManager idManager = new IdManager(net);
		addMainConfig(net, idManager);
		for (Map.Entry<ExtendedEntity ,Integer> e: statLevels.entrySet())
			addStatLevel(e.getKey(), net, e.getValue(), idManager);
	}
	
	private Element setupMetaAttributes(Element element) {
		Element metaAttributes = (Element)element.selectSingleNode("meta-attributes");
		if (metaAttributes == null) {
			metaAttributes = element.addElement("meta-attributes");
		}

		if (element.selectSingleNode("meta-attributes/meta-attribute[@configuration-name=\"" + configName + "\"]") != null)
			element.selectSingleNode("meta-attributes/meta-attribute[@configuration-name=\"" + configName + "\"]").detach();
		
		return metaAttributes;
	}
	
	private void addStatLevel(ExtendedEntity place, Element net, int level, IdManager idManager) {
		Element placeElement = NetHelper.getElement(net, place.getXmlId());
		
		Element metaAttributes = setupMetaAttributes(placeElement);
		
		if (place.getPlaceType() != PlaceType.PROBE)
			metaAttributes.addElement("meta-attribute")
			.addAttribute("xsi:type", "simqpn-place-configuration")
			.addAttribute("configuration-name", configName)
			.addAttribute("id", idManager.getNextId())
			.addAttribute("statsLevel", "" + level);
		else
			metaAttributes.addElement("meta-attribute")
			.addAttribute("xmlns:xsi", "http://www.w3.org/2001/XMLSchema-instance")
			.addAttribute("xsi:type", "simqpn-place-configuration")
			.addAttribute("configuration-name", configName)
			.addAttribute("id", idManager.getNextId())
			.addAttribute("statsLevel", "" + level);
		
		
		if (level >= 3) {
			for (Element colorRef: NetUtil.getColorReferencesForPlace(placeElement)) {
				Element refMetaAttributes = setupMetaAttributes(colorRef);
				refMetaAttributes.addElement("meta-attribute")
					.addAttribute("xsi:type", "simqpn-batch-means-color-configuration")
					.addAttribute("configuration-name", configName)
					.addAttribute("id", idManager.getNextId())
					.addAttribute("signLev", "0.05")
					.addAttribute("reqAbsPrc", "50")
					.addAttribute("reqRelPrc", "" + accuracy)
					.addAttribute("batchSize", "200")
					.addAttribute("minBatches", "51")
					.addAttribute("numBMeansCorlTested", "50")
					.addAttribute("bucketSize", "100.0")
					.addAttribute("maxBuckets", "1000")
					.addAttribute("queueSignLev", "0.05")
					.addAttribute("queueReqAbsPrc", "50")
					.addAttribute("queueReqRelPrc", "" + accuracy)
					.addAttribute("queueBatchSize", "200")
					.addAttribute("queueMinBatches", "60")
					.addAttribute("queueNumBMeansCorlTested", "50")
					.addAttribute("queueBucketSize", "100.0")
					.addAttribute("queueMaxBuckets", "1000");
			}
		}
	}
	
	private void addMainConfig(Element net, IdManager idManager) {
		Element mainNet = (Element)net.selectSingleNode("//net");
	
		Element metaAttributes = setupMetaAttributes(mainNet);
		
		metaAttributes.addElement("meta-attribute")
			.addAttribute("xmlns:xsi", "http://www.w3.org/2001/XMLSchema-instance")
			.addAttribute("configuration-name", configName)
			.addAttribute("xsi:type", "simqpn-configuration")
			.addAttribute("id", idManager.getNextId())
			.addAttribute("output-directory", outputDirectory)
			.addAttribute("ramp-up-length", "" + warmupPeriod)
			.addAttribute("scenario", "1")
			.addAttribute("time-before-initial-heart-beat", "" + timeBeforeInitialHeartBeat)
			.addAttribute("seconds-between-heart-beats", "" + secondsBetweenHeartBeats)
			.addAttribute("verbosity-level", "" + verbosity)
			.addAttribute("stopping-rule", stoppingRule)
			.addAttribute("time-between-stop-checks", "" + timeBetweenStopChecks)
			.addAttribute("seconds-between-stop-checks", "" + secondsBetweenStopChecks)
			.addAttribute("total-run-length", "" + totalRunLength);
	}
	
	public Double getAccuracy() {
		return accuracy;
	}
	
	public Double getWarmupPeriod() {
		return warmupPeriod;
	}
	public Double getTotalRunLength() {
		return totalRunLength;
	}

	public void setAccuracy(Double accuracy) {
		this.accuracy = accuracy;
	}

	public Double getSecondsBetweenStopChecks() {
		return secondsBetweenStopChecks;
	}
	
	private class IdManager {
		private static final String leadingString = "_";
		private long highestId;
		
		public IdManager(Element net) {
			highestId = findHighestId(net);
		}
		
		public String getNextId() {
			highestId++;
			return leadingString + highestId;
		}
		
	    private long findHighestId(Element element) {
	    	long highestId = 0;
	        for ( int i = 0, size = element.nodeCount(); i < size; i++ ) {
	            Node node = element.node(i);
	            if ( node instanceof Element ) {
	            	Element el = (Element) node;
	            	if (el.attributeValue("id") != null) {
	            		if (el.attributeValue("id").substring(1).matches("[0-9]+")) {
		            		long id = Long.parseLong(el.attributeValue("id").substring(1));
		            		if (id > highestId)
		            			highestId = id;
	            		}
	            	}
	            	long id = findHighestId( (Element) node );
	        		if (id > highestId)
	        			highestId = id;
	            }
	        }
	        return highestId;
	    }
	}
}
