package tools.descartes.dql.connector.simQPNV2;

import java.io.File;
import java.io.IOException;
import java.math.BigDecimal;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.LinkedList;
import java.util.List;
import java.util.Properties;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.dom4j.Element;
import org.dom4j.tree.DefaultAttribute;
import org.eclipse.emf.common.util.EList;

import de.tud.cs.qpe.model.NetHelper;
import de.tud.cs.simqpn.kernel.console.SimQPN;
import de.tud.cs.simqpn.kernel.monitor.ConsoleSimulatorProgress;
import extension.ExtendedDecimalResult;
import extension.ExtendedDoF;
import extension.ExtendedEntity;
import extension.ExtendedEntityMapping;
import extension.ExtendedProbe;
import extension.ExtensionUtil;
import extension.PlaceType;
import tools.descartes.dql.connector.PerformanceMetricsQueryConnector;
import tools.descartes.dql.connector.simQPNV2.DQLException;
import tools.descartes.dql.models.mapping.domain.DecimalResult;
import tools.descartes.dql.models.mapping.domain.DoF;
import tools.descartes.dql.models.mapping.domain.Probe;
import tools.descartes.dql.models.mapping.domain.Resource;
import tools.descartes.dql.models.mapping.domain.Service;
import tools.descartes.dql.models.mapping.mapping.EntityMapping;

public class PerformanceMetricsQueryConnectorImpl implements
		PerformanceMetricsQueryConnector {
	private static String resultFile = File.separator + "result.txt";
	private static String logFileName = File.separator + "log.txt";
	private static String modifiedNetFile = File.separator + "modifiedNet.qpe";
	private static String dqlFolderName = "DQL";
	private static String configurationName = "DQL-Configuration";
	private static double UNSET_ACCURACY = 0.95;
	private static String ACCURATE = "Accurate";
	private static double ACCURATE_ACCURACY = 0.95;
	private static String FAST = "Fast";
	private static double FAST_ACCURACY = 0.8;
	private String constraint = null;
	

	public List<EntityMapping> requestDoFMetrics(EntityMapping request) {
		ExtendedEntityMapping extendedRequest = ExtensionUtil.extendFullEntityMapping(request, new ModelStructureQueryConnectorImpl());

		Element net = FileUtil.loadXmlFromFile(request.getModelLocation());
		List<EntityMapping> results = new LinkedList<EntityMapping>();
		recursiveDoFWalk(extendedRequest, net, results);
		
		return results;
	}
	
	private void recursiveDoFWalk(ExtendedEntityMapping extendedRequest, 
			Element net, List<EntityMapping> results) {
		recursiveDoFWalk(extendedRequest, net, results, 0, 
				createFolderStructure(extendedRequest.getModelLocation(), true).getAbsolutePath());
	}
	
	private void recursiveDoFWalk(ExtendedEntityMapping extendedRequest, 
			Element net, List<EntityMapping> results, int counter, String path) {
		if (counter == extendedRequest.getDoF().size()) {
			path = path.substring(0, path.length()-1);
			results.add(startSimulation(path, net, extendedRequest));
		} else {
			ExtendedEntityMapping clonedRequest = extendedRequest.clone();
			ExtendedDoF currentDoF = (ExtendedDoF)clonedRequest.getDoF().get(counter);
			for (int i: currentDoF.getValues()) {
				currentDoF.getProbes().add(createDoFResult(i, currentDoF.getIdentifier()));
				Element changed = currentDoF.apply(net, i);
				recursiveDoFWalk(clonedRequest, changed, results, counter + 1, path + i + " ");
			}
		}
			
		
	}

	private DecimalResult createDoFResult(int i, String dofName) {
		ExtendedDecimalResult edr = new ExtendedDecimalResult();
		edr.setAccuracy(new BigDecimal("1.00"));
		edr.setValue(new BigDecimal(i));
		edr.setValid(true);
		edr.setMetricName(dofName);
		return edr;
	}

	public EntityMapping requestMetrics(EntityMapping request) {
		EntityMapping extendedRequest = ExtensionUtil.extendFullEntityMapping(request, new ModelStructureQueryConnectorImpl());
		
		File simulationFolder = createFolderStructure(request.getModelLocation(), false);
		Element net = FileUtil.loadXmlFromFile(request.getModelLocation());
		
		return startSimulation(simulationFolder.getAbsolutePath(), net, extendedRequest);
	}
	
	private File createFolderStructure(String modelLocation, boolean dof) {
		File qpn = new File(FileUtil.parsePath(modelLocation));
		File containingFolder = qpn.getParentFile();
		File dqlFolder = new File(containingFolder.getAbsolutePath() + File.separator 
				+ dqlFolderName);
		File modelFolder = new File(dqlFolder.getAbsoluteFile() + File.separator 
				+ qpn.getName().substring(0, qpn.getName().lastIndexOf(".")));
		Date now = new Date();
		SimpleDateFormat dateFormat = new SimpleDateFormat("yyyy MM dd hh mm ss SSSS");
		String time = dateFormat.format(now);
		if (dof)
			time += " DoF";
		File timeStamp = new File(modelFolder.getAbsoluteFile() + File.separator + time);
		timeStamp.mkdirs();
		timeStamp.mkdir();
		return timeStamp;
	}
	
	private EntityMapping startSimulation(String simulationFolderPath, Element net, EntityMapping request) {
		ConfigurationBuilder cb = new ConfigurationBuilder(configurationName);
		cb.setAccuracy(parseConstraint(constraint));
		setStatLevels(cb, request, net);
		cb.writeConfiguration(net);
		File results = createFile(simulationFolderPath + resultFile);
		File logfile = createFile(simulationFolderPath + logFileName);
		File modifiedNet = createFile(simulationFolderPath + modifiedNetFile);
		
		NetUtil.save(modifiedNet, net);

		SimQPN.runSimulator(modifiedNet.getAbsolutePath(), configurationName, results.getAbsolutePath(),
				getLoggingProperties(logfile), new ConsoleSimulatorProgress());
		
		parseResults(results, request);
		
		if (containsWarning(logfile)) {
			setResultsToFalse(request);
		}
		
		return request;
	}

	private void setResultsToFalse(EntityMapping request) {
		for (Resource r: request.getResources())
			for (Probe p: r.getProbes()) {
				DecimalResult res = (DecimalResult) p;
				res.setAccuracy(BigDecimal.valueOf(0.0));
				res.setValid(false);
			}
		for (Service s: request.getServices())
			for (Probe p: s.getProbes()) {
				DecimalResult res = (DecimalResult) p;
				res.setAccuracy(BigDecimal.valueOf(0.0));
				res.setValid(false);
			}
	}

	private boolean containsWarning(File logFile) {
		try {
			String toCheck = new String(Files.readAllBytes(Paths.get(logFile.getAbsolutePath())));
			return toCheck.contains("The required precision may not have been reached!");
		} catch (IOException e) {
			new DQLException("Internal Error: Could not read config File!");
		}
		return true;
	}
	
	private void parseResults(File resultsFile, EntityMapping request) {
		Element results = FileUtil.loadXmlFromFile(resultsFile.getAbsolutePath());
		
		List<ExtendedEntity> entities = new LinkedList<ExtendedEntity>();
		for (Resource e: request.getResources())
			entities.add((ExtendedEntity)e);
		for (Service e: request.getServices())
			entities.add((ExtendedEntity)e);
		for (ExtendedEntity ee: entities) {
			List<DecimalResult> decimalResults = new LinkedList<DecimalResult>();
			for (Probe p: ee.getProbes()) {
				String xPath = generateXPath(ee,(ExtendedProbe)p);
				BigDecimal value = BigDecimal.valueOf(Double.parseDouble(((DefaultAttribute)results
						.selectObject(xPath)).getValue()));
				decimalResults.add(((ExtendedProbe)p).createResult(value,
						BigDecimal.valueOf(parseConstraint(constraint))));
			}
			ee.getProbes().clear();
			ee.getProbes().addAll(decimalResults);
		}
	}

	private String generateXPath(ExtendedEntity ee, ExtendedProbe ep) {
		String xPath = "//simqpn-results/observed-element[@id=\"" 
							+ ee.getXmlId() + "\" and @type=\""
							+ getSimQPNPlaceType(ee.getPlaceType(), ep.isMetricForDepository())
							+ "\"]";
		if (ep.getColorName() != null)
			xPath += "/color[@name=\"" + ep.getColorName() + "\"]";
		
		xPath += "/metric[@type=\"" + ep.getSimQPNMetric().getType() + "\"]/@value";
		return xPath;
	}
	
	public String getSimQPNPlaceType(PlaceType placeType, boolean depository) {
		switch (placeType) {
		case QUEUE:
			return "queue";
		case QUEUEINGPLACE:
			if (depository)
				return "qplace:depository";
			else
				return "qplace:queue";
		case ORDINARYPLACE:
			return "place";
		case PROBE:
			return "probe";
		default:
			throw new DQLException();
		}
	}

	private Properties getLoggingProperties(File log) {
		Properties props = new Properties();
		props.setProperty("log4j.rootLogger", "WARN, File");
		props.setProperty("log4j.appender.File", "org.apache.log4j.RollingFileAppender");
		props.setProperty("log4j.appender.File.File", log.getAbsolutePath());
		props.setProperty("log4j.appender.File.layout", "org.apache.log4j.PatternLayout");
		props.setProperty("log4j.appender.File.maxFileSize", "1MB");
		return props;
	}
	
	private void setStatLevels(ConfigurationBuilder cb, EntityMapping request, Element net) {
		for (Element p: NetHelper.listAllPlaces(net))
			cb.addStatLevel(new ExtendedEntity(NetUtil.getIdOfElement(p), PlaceType.ORDINARYPLACE,
					"", null), 0, net);
			
		List<ExtendedEntity> entities = new LinkedList<ExtendedEntity>();
		for (Resource e: request.getResources())
			entities.add((ExtendedEntity)e);
		for (Service e: request.getServices())
			entities.add((ExtendedEntity)e);
		for (ExtendedEntity ee: entities) {
			for (Probe p: ee.getProbes())
				cb.addStatLevel(ee, ((ExtendedProbe)p).getStatLevel(), net);
		}
	}

	public double parseConstraint(String constraint) {
		if (constraint == null)
			return UNSET_ACCURACY;
		if (constraint.equals(ACCURATE))
			return ACCURATE_ACCURACY;
		if (constraint.equals(FAST))
			return FAST_ACCURACY;
		Pattern pattern = Pattern.compile("[Aa]ccuracy[ ]?=[ ]?([0-9]+.?[0-9]+)");
		Matcher m = pattern.matcher(constraint);
		if (m.matches()) {
			double accuracy = Double.parseDouble(m.group(1));
			if (accuracy > 1.0)
				throw new DQLException("Accuracy has to be below 1.0");
			return accuracy;
		}
		throw new DQLException(constraint + " is not a valid Constraint!");
	}
	
	public File createFile(String filePath) {
		File f = new File(filePath);
		try {
			f.createNewFile();
		} catch (IOException e) {
			new DQLException("Was unable to create required Folders/Files!");
		}
		return f;
	}
	
	public void reset() {
		constraint = null;
	}

	public String getBundleProperty(String key) {
		return Activator.getContext().getProperty(key);
	}

	public String getConstraint() {
		return constraint;
	}

	public void setConstraint(String constraint) {
		this.constraint = constraint;
	}
	
	
	
	
	
	
	
	
	//SUPPORTS STUFF
	
	
	public boolean supportsConstraintRequests() {
		return true;
	}

	public boolean supportsDoFRequests() {
		return true;
	}

	public boolean supportsMetricRequests() {
		return true;
	}

	public boolean supportsObservationPeriods() {
		return false;
	}
	

	
	
	
	
	
	
	
	
	
	
	
	//USELESS STUBS

	public String getDoFConstraint() {
		return null;
	}

	public String getDoFExplorationStrategy() {
		return null;
	}

	public Properties getDoFExplorationStrategyProperties() {
		return null;
	}

	public List<String> getSupportedConstraintNames() {
		return null;
	}

	public List<String> getSupportedDoFConstraintNames() {
		return null;
	}

	public List<String> getSupportedExplorationStrategyNames() {
		return null;
	}

	public List<String> getSupportedTimeUnits() {
		return null;
	}
	
	public void setDoFConstraint(String doFConstraint) {
	}

	public void setDoFExplorationStrategy(String strategy, Properties properties) {
	}

	public void setObservationPeriod(Date start, Date end, long sampleSize) {
	}

	public void setObservationPeriod(long amount, String unit, long sampleSize) {
	}

	public void setObservationReference(String type, String identifier,
			long sampleSize) {
	}

}